@@ -9,7 +9,7 @@
 
 /*
   Copyright (C) 2002 Dario Abatianni <eisfuchs@tigress.com>
-  Copyright (C) 2005-2007 Peter Simonsson <psn@linux.se>
+  Copyright (C) 2005-2016 Peter Simonsson <peter.simonsson@gmail.com>
   Copyright (C) 2006-2010 Eike Hein <hein@kde.org>
   Copyright (C) 2004-2011 Eli Mackenzie <argonel@gmail.com>
 */
@@ -128,8 +128,6 @@ IRCView::IRCView(QWidget* parent) : QTextBrowser(parent), m_rememberLine(0), m_l
     setTextInteractionFlags(Qt::TextBrowserInteraction);
     viewport()->setMouseTracking(true);
 
-    if (Preferences::self()->useParagraphSpacing()) enableParagraphSpacing();
-
     //HACK to workaround an issue with the QTextDocument
     //doing a relayout/scrollbar over and over resulting in 100%
     //proc usage. See bug 215256
@@ -142,6 +140,25 @@ IRCView::~IRCView()
 {
 }
 
+void IRCView::increaseFontSize()
+{
+    QFont newFont(font());
+    newFont.setPointSize(font().pointSize() + 1);
+    setFont(newFont);
+}
+
+void IRCView::decreaseFontSize()
+{
+    QFont newFont(font());
+    newFont.setPointSize(font().pointSize() - 1);
+    setFont(newFont);
+}
+
+void IRCView::resetFontSize()
+{
+    setFont(Preferences::self()->textFont());
+}
+
 void IRCView::setServer(Server* newServer)
 {
     if (m_server == newServer)
@@ -170,34 +187,30 @@ void IRCView::findPreviousText()
     emit doSearchPrevious();
 }
 
-bool IRCView::search(const QString& pattern, bool caseSensitive, bool wholeWords, bool forward, bool fromCursor)
+bool IRCView::search(const QString& pattern, QTextDocument::FindFlags flags, bool fromCursor)
 {
     if (pattern.isEmpty())
         return true;
 
     m_pattern       = pattern;
-    m_forward       = forward;
-    m_searchFlags = 0;
-    if (caseSensitive)
-        m_searchFlags |= QTextDocument::FindCaseSensitively;
-    if (wholeWords)
-        m_searchFlags |= QTextDocument::FindWholeWords;
+    m_searchFlags = flags;
+
     if (!fromCursor)
-        m_forward ? moveCursor(QTextCursor::Start) : moveCursor(QTextCursor::End);
+        moveCursor(QTextCursor::End);
+    else
+        moveCursor(QTextCursor::StartOfWord); // Do this to that if possible the same position is kept when changing search options
 
     return searchNext();
 }
 
 bool IRCView::searchNext(bool reversed)
 {
-    bool fwd = (reversed ? !m_forward : m_forward);
-    if (fwd) {
-        m_searchFlags &= ~QTextDocument::FindBackward;
-    }
-    else {
-        m_searchFlags |= QTextDocument::FindBackward;
-    }
-    return find(m_pattern, m_searchFlags);
+    QTextDocument::FindFlags flags = m_searchFlags;
+
+    if(!reversed)
+        flags |= QTextDocument::FindBackward;
+
+    return find(m_pattern, flags);
 }
 
 class IrcViewMimeData : public QMimeData
@@ -494,14 +507,18 @@ Burr* IRCView::appendLine(IRCView::ObjectFormats type)
 
 // Other stuff
 
-void IRCView::enableParagraphSpacing() {}
-
 void IRCView::updateAppearance()
 {
+    QFont newFont;
+    const int fontSize = font().pointSize();
+
     if (Preferences::self()->customTextFont())
-        setFont(Preferences::self()->textFont());
+        newFont = Preferences::self()->textFont();
     else
-        setFont(QFontDatabase::systemFont(QFontDatabase::GeneralFont));
+        newFont = QFontDatabase::systemFont(QFontDatabase::GeneralFont);
+
+    newFont.setPointSize(fontSize);
+    setFont(newFont);
 
     setVerticalScrollBarPolicy(Preferences::self()->showIRCViewScrollBar() ? Qt::ScrollBarAlwaysOn : Qt::ScrollBarAlwaysOff);
 
@@ -525,10 +542,67 @@ void IRCView::updateAppearance()
     viewport()->setPalette(p);
 }
 
+bool IRCView::dateRtlDirection()
+{
+    // Keep format synced with IRCView::timeStamp
+    return QLocale().toString(QDate::currentDate(), QLocale::ShortFormat).isRightToLeft();
+}
+
+// To minimize the use of bidi marks, for cases below, some bidi marks are
+// needed.
+// * left aligned lines in LTR locales, and
+// * right aligned lines in RTL locales
+//
+// First, check if the direction of the message is the same as the
+// direction of the timestamp, if not, then add a mark depending on
+// message's direction, so that timestamp don't be first strong character.
+//
+// If we have a LTR label, and the message is right-aligned, we prepend
+// it with LRM to look correct (check nickname case below), and then append
+// it with LRM also and then a RLM to preserve the direction of the
+// right-aligned line.
+//
+// Later, if the message is RTL, nicknames like "_nick]" will appear
+// as "[nick_".
+// First, add a LRM mark to make underscore on the left, next add the
+// nickname, and then another LRM mark. Since we use RTL/LTR count, the
+// message may start with a LTR word, and appear to the right of the
+// nickname. That's why we add a RLM mark before the nick to force it
+// appearing on left.
+QString IRCView::formatFinalLine(bool rtl, QString lineColor, QString label, QString nickLine, QString nickStar, QString text)
+{
+    // Nick correctly displayed: <_nick]>
+    QString line;
+    // It's right-aligned under LTR locale, or left-aligned under RTL locale
+    if (!rtl == dateRtlDirection())
+        line += (rtl ? RLM : LRM);
+    if (!label.isEmpty()) { // Label correctly displayed: [_label.]
+        if (rtl) line += LRM; // [.label_] -> [._label]
+        line += "<font color=\"" + lineColor + "\"><b>[</b>%4<b>]</b></font>";
+        if (!label.isRightToLeft() == rtl)
+            line += LRM + RLM; // [._label] -> [_label.]
+    }
+    line += "<font color=\"" + lineColor + "\">%1";
+    if (!nickStar.isEmpty()) // Used for [timeStamp] * nick action
+        line += nickStar;
+    if (rtl) line += LRM; // <[nick_> -> <[_nick]>
+    line += nickLine;
+    if (rtl)
+    {
+        line += LRM; // <[_nick]> -> <_nick]>
+        // It might start with an English word, but it's RTL because of counting
+        if (!text.isEmpty() && !text.isRightToLeft())
+            line += RLM; // ARABIC_TEXT <_nick]> Hi -> ARABIC_TEXT Hi <_nick]>
+    }
+    if (text.isEmpty()) line += "</font>";
+    else line += " %3</font>";
+
+    return line;
+}
 
 // Data insertion
 
-void IRCView::append(const QString& nick, const QString& message)
+void IRCView::append(const QString& nick, const QString& message, const QHash<QString, QString> &messageTags, const QString& label)
 {
     QString channelColor = Preferences::self()->color(Preferences::ChannelMessage).name();
 
@@ -541,22 +615,16 @@ void IRCView::append(const QString& nick, const QString& message)
     QString line;
     bool rtl = (dir == QChar::DirR);
 
-    if(rtl)
-    {
-        line = RLE;
-        line += LRE;
-        line += "<font color=\"" + channelColor + "\">" + nickLine +" %1" + PDF + RLM + " %3</font>";
-    }
-    else
-    {
-        if (!QApplication::isLeftToRight())
-            line += LRE;
+    // Normal chat lines
+    // [timestamp] <nickname> chat message
+    line = formatFinalLine(rtl, channelColor, label, nickLine, QString(), text);
+    line = line.arg(timeStamp(messageTags, rtl), nick, text);
 
-        line += "<font color=\"" + channelColor + "\">%1" + nickLine + " %3</font>";
+    if (!label.isEmpty())
+    {
+        line = line.arg(label);
     }
 
-    line = line.arg(timeStamp(), nick, text);
-
     emit textToLog(QString("<%1>\t%2").arg(nick, message));
 
     doAppend(line, rtl);
@@ -568,7 +636,15 @@ void IRCView::appendRaw(const QString& message, bool self)
         : Preferences::self()->color(Preferences::ServerMessage);
     m_tabNotification = Konversation::tnfNone;
 
-    QString line = QString(timeStamp() + " <font color=\"" + color.name() + "\">" + message + "</font>");
+    // Raw log is always left-aligned
+    // [timestamp] << server line
+    // If the timedate string is RTL, prepend a LTR mark to force the direction
+    // to be LTR, as the datetime string is already returned as it's for a
+    // left-aligned line.
+
+    QString line;
+    if (dateRtlDirection()) line += LRM;
+    line += (timeStamp(QHash<QString, QString>(), false) + " <font color=\"" + color.name() + "\">" + message + "</font>");
 
     doAppend(line, false, self);
 }
@@ -578,12 +654,15 @@ void IRCView::appendLog(const QString & message)
     QColor channelColor = Preferences::self()->color(Preferences::ChannelMessage);
     m_tabNotification = Konversation::tnfNone;
 
+    // Log view is plain log files.
+    // Direction will be depending on the logfile line direction.
+
     QString line("<font color=\"" + channelColor.name() + "\">" + message + "</font>");
 
-    doRawAppend(line, !QApplication::isLeftToRight());
+    doRawAppend(line, message.isRightToLeft());
 }
 
-void IRCView::appendQuery(const QString& nick, const QString& message, bool inChannel)
+void IRCView::appendQuery(const QString& nick, const QString& message, const QHash<QString, QString> &messageTags, bool inChannel)
 {
     QString queryColor=Preferences::self()->color(Preferences::QueryMessage).name();
 
@@ -596,21 +675,10 @@ void IRCView::appendQuery(const QString& nick, const QString& message, bool inCh
     QString text(filter(message, queryColor, nick, true, true, false, &dir));
     bool rtl = (dir == QChar::DirR);
 
-    if(rtl)
-    {
-        line = RLE;
-        line += LRE;
-        line += "<font color=\"" + queryColor + "\">" + nickLine + " %1" + PDF + RLM + " %3</font>";
-    }
-    else
-    {
-        if (!QApplication::isLeftToRight())
-            line += LRE;
-
-        line += "<font color=\"" + queryColor + "\">%1 " + nickLine + " %3</font>";
-    }
-
-    line = line.arg(timeStamp(), nick, text);
+    // Private chat lines
+    // [timestamp] <nickname> chat message
+    line = formatFinalLine(rtl, queryColor, QString(), nickLine, QString(), text);
+    line = line.arg(timeStamp(messageTags, rtl), nick, text);
 
     if (inChannel) {
         emit textToLog(QString("<-> %1>\t%2").arg(nick, message));
@@ -621,19 +689,19 @@ void IRCView::appendQuery(const QString& nick, const QString& message, bool inCh
     doAppend(line, rtl);
 }
 
-void IRCView::appendChannelAction(const QString& nick, const QString& message)
+void IRCView::appendChannelAction(const QString& nick, const QString& message, const QHash<QString, QString> &messageTags)
 {
     m_tabNotification = Konversation::tnfNormal;
-    appendAction(nick, message);
+    appendAction(nick, message, messageTags);
 }
 
-void IRCView::appendQueryAction(const QString& nick, const QString& message)
+void IRCView::appendQueryAction(const QString& nick, const QString& message, const QHash<QString, QString> &messageTags)
 {
     m_tabNotification = Konversation::tnfPrivate;
-    appendAction(nick, message);
+    appendAction(nick, message, messageTags);
 }
 
-void IRCView::appendAction(const QString& nick, const QString& message)
+void IRCView::appendAction(const QString& nick, const QString& message, const QHash<QString, QString> &messageTags)
 {
     QString actionColor = Preferences::self()->color(Preferences::ActionMessage).name();
 
@@ -643,16 +711,17 @@ void IRCView::appendAction(const QString& nick, const QString& message)
 
     if (message.isEmpty())
     {
-        if (!QApplication::isLeftToRight())
-            line += LRE;
+        // No text to check direction. Better to check last line, if it's RTL,
+        // treat it as that.
+        QTextCursor formatCursor(document()->lastBlock());
+        bool rtl = (formatCursor.blockFormat().alignment().testFlag(Qt::AlignRight));
 
-        line += "<font color=\"" + actionColor + "\">%1 * " + nickLine + "</font>";
-
-        line = line.arg(timeStamp(), nick);
+        line = formatFinalLine(rtl, actionColor, QString(), nickLine, QString(" * "), QString());
+        line = line.arg(timeStamp(messageTags, rtl), nick);
 
         emit textToLog(QString("\t * %1").arg(nick));
 
-        doAppend(line, false);
+        doAppend(line, rtl);
     }
     else
     {
@@ -660,21 +729,10 @@ void IRCView::appendAction(const QString& nick, const QString& message)
         QString text(filter(message, actionColor, nick, true,true, false, &dir));
         bool rtl = (dir == QChar::DirR);
 
-        if (rtl)
-        {
-            line = RLE;
-            line += LRE;
-            line += "<font color=\"" + actionColor + "\">" + nickLine + " * %1" + PDF + " %3</font>";
-        }
-        else
-        {
-            if (!QApplication::isLeftToRight())
-                line += LRE;
-
-            line += "<font color=\"" + actionColor + "\">%1 * " + nickLine + " %3</font>";
-        }
-
-        line = line.arg(timeStamp(), nick, text);
+        // Actions line
+        // [timestamp] * nickname action
+        line = formatFinalLine(rtl, actionColor, QString(), nickLine, QString(" * "), text);
+        line = line.arg(timeStamp(messageTags, rtl), nick, text);
 
         emit textToLog(QString("\t * %1 %2").arg(nick, message));
 
@@ -682,7 +740,7 @@ void IRCView::appendAction(const QString& nick, const QString& message)
     }
 }
 
-void IRCView::appendServerMessage(const QString& type, const QString& message, bool parseURL)
+void IRCView::appendServerMessage(const QString& type, const QString& message, const QHash<QString, QString> &messageTags, bool parseURL)
 {
     QString serverColor = Preferences::self()->color(Preferences::ServerMessage).name();
     m_tabNotification = Konversation::tnfControl;
@@ -698,30 +756,24 @@ void IRCView::appendServerMessage(const QString& type, const QString& message, b
     QString line;
     QChar::Direction dir;
     QString text(filter(message, serverColor, 0 , true, parseURL, false, &dir));
-    bool rtl = (dir == QChar::DirR);
+    // Server text may be translated strings. It's not user input: treat with first strong.
+    bool rtl = text.isRightToLeft();
 
-    if(rtl)
-    {
-        line = RLE;
-        line += LRE;
-        line += "<font color=\"" + serverColor + "\"" + fixed + "><b>[</b>%2<b>]</b> %1" + PDF + " %3</font>";
-    }
-    else
-    {
-        if (!QApplication::isLeftToRight())
-            line += LRE;
-
-        line += "<font color=\"" + serverColor + "\"" + fixed + ">%1 <b>[</b>%2<b>]</b> %3</font>";
-    }
-
-    line = line.arg(timeStamp(), type, text);
+    // It's right-aligned under LTR locale, or left-aligned under RTL locale
+    if (!rtl == dateRtlDirection())
+        line += (rtl ? RLM : LRM);
+    line += "<font color=\"" + serverColor + "\"" + fixed + ">%1 <b>[</b>%2<b>]</b>";
+    if (!rtl == type.isRightToLeft())
+        line += (rtl ? RLM : LRM); // [50 [ARABIC_TEXT users -> [ARABIC_TEXT] 50 users
+    line += " %3</font>";
+    line = line.arg(timeStamp(messageTags, rtl), type, text);
 
     emit textToLog(QString("%1\t%2").arg(type, message));
 
     doAppend(line, rtl);
 }
 
-void IRCView::appendCommandMessage(const QString& type,const QString& message, bool parseURL, bool self)
+void IRCView::appendCommandMessage(const QString& type, const QString& message, const QHash<QString, QString> &messageTags, bool parseURL, bool self)
 {
     QString commandColor = Preferences::self()->color(Preferences::CommandMessage).name();
     QString prefix="***";
@@ -742,23 +794,14 @@ void IRCView::appendCommandMessage(const QString& type,const QString& message, b
     QString line;
     QChar::Direction dir;
     QString text(filter(message, commandColor, 0, true, parseURL, self, &dir));
-    bool rtl = (dir == QChar::DirR);
+    // Commands are translated and contain LTR IP addresses. Treat with first strong.
+    bool rtl = text.isRightToLeft();
 
-    if(rtl)
-    {
-        line = RLE;
-        line += LRE;
-        line += "<font color=\"" + commandColor + "\">%2 %1" + PDF + " %3</font>";
-    }
-    else
-    {
-        if (!QApplication::isLeftToRight())
-            line += LRE;
-
-        line += "<font color=\"" + commandColor + "\">%1 %2 %3</font>";
-    }
-
-    line = line.arg(timeStamp(), prefix, text);
+    // It's right-aligned under LTR locale, or left-aligned under RTL locale
+    if (!rtl == dateRtlDirection())
+        line += (rtl ? RLM : LRM);
+    line += "<font color=\"" + commandColor + "\">%1 %2 %3</font>";
+    line = line.arg(timeStamp(messageTags, rtl), prefix, text);
 
     emit textToLog(QString("%1\t%2").arg(type, message));
 
@@ -790,22 +833,36 @@ void IRCView::appendBacklogMessage(const QString& firstColumn,const QString& raw
     QString line;
     QChar::Direction dir;
     QString text(filter(message, backlogColor, NULL, false, false, false, &dir));
-    bool rtl = (dir == QChar::DirR);
+    bool rtl = nick.startsWith('|') ? text.isRightToLeft() : (dir == QChar::DirR);
 
-    if(rtl)
-    {
-        line = RLE;
-        line += LRE;
-        line += "<font color=\"" + backlogColor + "\">%2 %1" + PDF + " %3</font>";
-    }
+    // It's right-aligned under LTR locale, or left-aligned under RTL locale
+    if (!rtl == time.isRightToLeft())
+        line += (rtl ? RLM : LRM);
+
+    line += "<font color=\"" + backlogColor + "\">";
+    // Prepend and append timestamp's correct bidi mark if the time and text
+    // directions are different.
+    if (rtl == time.isRightToLeft())
+        line += "%1";
     else
-    {
-        if (!QApplication::isLeftToRight())
-            line += LRE;
+        line += (time.isRightToLeft() ? RLM+"%1"+RLM : LRM+"%1"+LRM);
 
-        line += "<font color=\"" + backlogColor + "\">%1 %2 %3</font>";
+    // Partially copied from IRCView::formatFinalLine
+    if (rtl)
+    {
+        // Return back to the normal direction after setting mark
+        if (!rtl == time.isRightToLeft()) line += (!time.isRightToLeft() ? RLM : LRM);
+        line += LRM; // <[nick_> -> <[_nick]>
+    }
+    line += "%2";
+    if (rtl)
+    {
+        line += LRM; // <[_nick]> -> <_nick]>
+        if (!text.isRightToLeft())
+            line += RLM; // ARABIC_TEXT <_nick]> Hi -> ARABIC_TEXT Hi <_nick]>
     }
 
+    line +=  " %3</font>";
     line = line.arg(time, nick, text);
 
     doAppend(line, rtl);
@@ -860,30 +917,39 @@ void IRCView::doRawAppend(const QString& newLine, bool rtl)
     QTextCursor formatCursor(document()->lastBlock());
     QTextBlockFormat format = formatCursor.blockFormat();
 
-    format.setAlignment(rtl ? Qt::AlignRight : Qt::AlignLeft);
+    format.setAlignment(Qt::AlignAbsolute|(rtl ? Qt::AlignRight : Qt::AlignLeft));
     formatCursor.setBlockFormat(format);
 }
 
-QString IRCView::timeStamp()
+QString IRCView::timeStamp(QHash<QString, QString> messageTags, bool rtl)
 {
     if(Preferences::self()->timestamping())
     {
-        QTime time = QTime::currentTime();
+        QDateTime serverTime;
+
+        if (messageTags.contains(QStringLiteral("time"))) // If it exists use the supplied server time.
+            serverTime = QDateTime::fromString(messageTags[QStringLiteral("time")], Qt::ISODate).toLocalTime();
+
+        QTime time = serverTime.isValid() ? serverTime.time() : QTime::currentTime();
         QString timeColor = Preferences::self()->color(Preferences::Time).name();
         QString timeFormat = Preferences::self()->timestampFormat();
         QString timeString;
 
+        bool dateRtl = dateRtlDirection();
+
         if(!Preferences::self()->showDate())
         {
             timeString = QString(QLatin1String("<font color=\"") + timeColor + QLatin1String("\">[%1]</font> ")).arg(time.toString(timeFormat));
         }
         else
         {
-            QDate date = QDate::currentDate();
+            QDate date = serverTime.isValid() ? serverTime.date() : QDate::currentDate();
             timeString = QString("<font color=\"" +
-                timeColor + "\">[%1 %2]</font> ")
-                    .arg(QLocale().toString(date, QLocale::ShortFormat),
-                         time.toString(timeFormat));
+                timeColor + "\">[%1%2 %3%4]</font> ")
+                    .arg((dateRtl==rtl) ? QString() : (dateRtl ? RLM : LRM),
+                        QLocale().toString(date, QLocale::ShortFormat),
+                        time.toString(timeFormat),
+                         (dateRtl==rtl) ? QString() : (!dateRtl ? RLM : LRM));
         }
 
         return timeString;
@@ -894,7 +960,7 @@ QString IRCView::timeStamp()
 
 QString IRCView::createNickLine(const QString& nick, const QString& defaultColor, bool encapsulateNick, bool privMsg)
 {
-    QString nickLine = "%2";
+    QString nickLine ="%2";
     QString nickColor;
 
     if (Preferences::self()->useColoredNicks())
@@ -1018,7 +1084,7 @@ QString IRCView::filter(const QString& line, const QString& defaultColor, const
             while (it.hasNext())
             {
                 highlight = it.next();
-                QStringList highlightChatWindowList = highlight->getChatWindowList();
+                highlightChatWindowList = highlight->getChatWindowList();
                 if (highlightChatWindowList.isEmpty() ||
                     highlightChatWindowList.contains(m_chatWin->getName(), Qt::CaseInsensitive))
                 {
@@ -1595,7 +1661,8 @@ QString IRCView::closeToTagString(TextHtmlData* data, const QString& _tag)
     }
 
     // reopen relevant tags
-    ret += openTags(data, i);
+    if (i > -1)
+        ret += openTags(data, i);
 
     return ret;
 }
@@ -1603,7 +1670,8 @@ QString IRCView::closeToTagString(TextHtmlData* data, const QString& _tag)
 QString IRCView::openTags(TextHtmlData* data, int from)
 {
     QString ret, tag;
-    int i = from;
+    int i = from > -1 ? from : 0;
+
     for ( ;  i < data->openHtmlTags.count(); ++i)
     {
         tag = data->openHtmlTags.at(i);
@@ -1954,6 +2022,17 @@ void IRCView::mousePressEvent(QMouseEvent* ev)
     QTextBrowser::mousePressEvent(ev);
 }
 
+void IRCView::wheelEvent(QWheelEvent *ev)
+{
+    if(ev->modifiers()==Qt::ControlModifier)
+    {
+        if(ev->delta() < 0) decreaseFontSize();
+        if(ev->delta() > 0) increaseFontSize();
+    }
+
+    QTextBrowser::wheelEvent(ev);
+}
+
 void IRCView::mouseReleaseEvent(QMouseEvent *ev)
 {
     if (ev->button() == Qt::LeftButton)
@@ -2003,6 +2082,10 @@ void IRCView::openLink(const QUrl& url)
     QString link(url.toString());
     // HACK Replace " " with %20 for channelnames, NOTE there can't be 2 channelnames in one link
     link = link.replace (' ', "%20");
+    // HACK Handle pipe as toString doesn't seem to decode that correctly
+    link = link.replace ("%7C", "|");
+    // HACK Handle ` as toString doesn't seem to decode that correctly
+    link = link.replace ("%60", "`");
 
     if (!link.isEmpty() && !link.startsWith('#'))
         Application::openUrl(url.toEncoded());
